#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <pthread.h>
#include "is_prime.h"

typedef struct {
	int capacity, num_found;
	int *data;

	// TODO: add a mutex for synchronization
	pthread_mutex_t lock;
} PrimeList ;

// Create a PrimeList with enough capacity to store given
// number of primes.
PrimeList *primelist_alloc(int capacity)
{
	PrimeList *list = malloc(sizeof(PrimeList));
	list->capacity = capacity;
	list->num_found = 0;
	list->data = malloc(capacity * sizeof(int));

	// TODO: initialize the mutex
	pthread_mutex_init(&list->lock, NULL);

	return list;
}

// Use this function to add a prime to the PrimeList.
// This function must be thread-safe!
// Use the PrimeList's mutex to guard the critical section
// (the code that will access the PrimeList's fields).
void primelist_append(PrimeList *list, int prime)
{
	// TODO
	pthread_mutex_lock(&list->lock);

	// make sure there is room to store one more value
	if (list->num_found < list->capacity) {
		// append the new value to the end of the list
		list->data[list->num_found] = prime;
		// now we have one more value
		list->num_found++;
	}

	pthread_mutex_unlock(&list->lock);
}

struct Work {
	PrimeList *list;
	int thread_num;
	int start, end;
};

void *worker(void *arg)
{
	struct Work *w = arg;

	printf("Thread %i: range is %i..%i\n", w->thread_num, w->start, w->end);
	for (int i = w->start; i < w->end; i++) {
		if (is_prime(i)) {
			primelist_append(w->list, i);
		}
	}

	return NULL;
}

int main(void)
{
	int min, max;
	int num_threads;

	printf("Min: ");
	scanf("%i", &min);
	printf("Max: ");
	scanf("%i", &max);
	printf("Number of threads: ");
	scanf("%i", &num_threads);

	int est_num_primes = (int) ((max / log(max)) * 1.4);
	printf("Estimated number of primes is %i\n", est_num_primes);

	// Use this data structure to store all of the primes discovered
	PrimeList *list = primelist_alloc(est_num_primes);

	// TODO: use threads to find all primes in range, add them to the PrimeList
	int n = (max-min)+1;
	int chunk_size = n/num_threads;

	struct Work work_list[num_threads];

	for (int i = 0; i < num_threads; i++) {
		work_list[i].list = list;
		work_list[i].thread_num = i;
		work_list[i].start = min + (i*chunk_size);
		work_list[i].end = min + (i+1)*chunk_size;
		if (i == num_threads-1) {
			// last thread gets excess
			work_list[i].end += n%num_threads;
		}
	}

	// start threads
	pthread_t thr[num_threads];
	for (int i = 0; i < num_threads; i++) {
		pthread_create(&thr[i], NULL, &worker, &work_list[i]);
	}

	// wait for threads to exit
	for (int i = 0; i < num_threads; i++) {
		pthread_join(thr[i], NULL);
		printf("Thread %i finished\n", i);
	}

	// TODO: print out all primes found in range (once all threads have completed)
	printf("%i primes found\n", list->num_found);
	for (int i = 0; i < list->num_found; i++) {
		printf("%i\n", list->data[i]);
	}

	return 0;
}
